# -*- coding: utf-8 -*-
"""
Created on Mon Nov 11 11:36:25 2024

@author: Giovanni
"""

import numpy as np
import matplotlib.pyplot as plt

def PCA_reconstruction(train_data_name, test_data_name, perc_var, plot=True):
    # Load train and test data
    train_data = np.loadtxt(train_data_name)
    test_data = np.loadtxt(test_data_name)

    # Get data dimensions
    N, n = train_data.shape

    # Data pretreatment: Normalize training data (mean = 0, std = 1 for each signal)
    m_data = np.mean(train_data, axis=0)
    std_data = np.std(train_data, axis=0)
    train_data_n = (train_data - m_data) / std_data

    # Training Phase: Compute covariance matrix
    V = np.cov(train_data_n, rowvar=False)

    # Compute eigenvalues and eigenvectors
    eig_values, P_rev = np.linalg.eigh(V)
    # Order eigenvalues and eigenvectors from largest to smallest
    idx = np.argsort(eig_values)[::-1]
    eig_values = eig_values[idx]
    P = P_rev[:, idx]

    # Determine number of principal components to retain based on variance threshold
    sum_eig = np.sum(eig_values)
    n_PC = 0
    sum_var = 0
    while sum_var < perc_var:
        sum_var += eig_values[n_PC] / sum_eig
        n_PC += 1
        if n_PC == n:  # Avoid numerical issues
            sum_var = 1

    # Retain only the first n_PC principal components
    P_lambda = P[:, :n_PC]

    # Test Phase: Normalize test data using training mean and std deviation
    N_test = test_data.shape[0]
    test_data_n = (test_data - m_data) / std_data

    # Reconstruct test data in transformed basis and denormalize
    test_data_rec_n = test_data_n @ P_lambda @ P_lambda.T
    test_data_rec = test_data_rec_n * std_data + m_data

    # Compute the root mean square error (RMSE)
    rmse = np.sqrt(np.sum((test_data_rec - test_data) ** 2) / (n * N_test))

    # Plot the results
    signal_names = ['T1', 'T2', 'T3', 'T4', 'T5', 'T6']
    measure_error = 0.3
    if plot:
        for i in range(n):
            plt.figure()
    
            # Original and reconstructed signals
            plt.subplot(2, 1, 1)
            plt.plot(test_data[:, i], 'b', label='true value')
            plt.plot(test_data_rec[:, i], 'r', label='reconstruction')
            plt.xlabel(signal_names[i])
            plt.legend()
    
            # Residuals with error bounds
            plt.subplot(2, 1, 2)
            residuals = test_data[:, i] - test_data_rec[:, i]
            plt.plot(residuals, 'r', label='residual')
            upper_error = np.full(N_test, 3 * measure_error)
            lower_error = -upper_error
            plt.plot(upper_error, 'g--')
            plt.plot(lower_error, 'g--')
            plt.ylabel('residual')
            plt.xlabel(signal_names[i])
            plt.axis([0, N_test, -10 * measure_error, 10 * measure_error])
    
            plt.tight_layout()
            plt.show()

    return test_data_rec, n_PC, rmse
